import base64
import json
from datetime import timedelta, datetime

from crispy_forms.bootstrap import StrictButton
from date_offset.date_offset import DateOffset
from django.apps import apps
from django.db import ProgrammingError, DataError
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.utils.safestring import mark_safe
from django.views.generic import TemplateView
from django_datatables.datatables import DatatableTable
from django_menus.menu import MenuItem
from django_modals.forms import CrispyForm

from advanced_report_builder.columns import ReportBuilderDateColumn
from advanced_report_builder.exceptions import ReportError
from advanced_report_builder.globals import NUMBER_FIELDS, ANNOTATION_VALUE_FUNCTIONS
from advanced_report_builder.models import ReportType
from advanced_report_builder.utils import split_slug, get_field_details, split_attr
from advanced_report_builder.views.report import ReportBase
from advanced_report_builder.views.report_utils_mixin import ReportUtilsMixin
from advanced_report_builder.views.targets.utils import get_target_value


class ChartJSTable(DatatableTable):

    def __init__(self, *args, **kwargs):
        pk = kwargs.pop('pk', None)
        self.axis_scale = kwargs.pop('axis_scale', None)
        self.targets = kwargs.pop('targets', None)
        super().__init__(*args, **kwargs)
        if pk:
            self.filter['pk'] = pk

    def model_table_setup(self):
        try:
            targets = []
            try:
                data = self.get_table_array(self.kwargs.get('request'), self.get_query())
            except DataError:
                data = [['N/A']]
            if self.targets is not None:
                targets = self.targets.all()
                if targets and len(data) > 0:
                    targets = self.process_data_structure_target(targets=targets, data=data)
            return mark_safe(json.dumps({'initsetup': json.loads(self.col_def_str()),
                                         'data': data,
                                         'targets': targets,
                                         'row_titles': self.all_titles(),
                                         'table_id': self.table_id}))
        except ProgrammingError as e:
            raise ReportError(e)

    def process_data_structure_target(self, targets, data):
        results = []
        for target in targets:
            new_data_structure = []
            for data_dict in data:
                target_value = self.process_target_results(data_dict=data_dict, target=target)
                new_data_structure.append(target_value)
            label = target.name + ' Target'
            colour = '#' + target.colour
            results.append({'label': label,
                            'data': new_data_structure,
                            'borderWidth': 1,
                            'backgroundColor': [colour for _ in range(len(new_data_structure))],
                            'borderColor': '#' + target.colour})
        return results

    @staticmethod
    def process_target_results(data_dict, target):
        """
        Get the correct target data from from the target.
        :param data_dict:
        :param target:
        :return:
        """
        start_date = datetime.strptime(data_dict[0], '%Y-%m-%d').date()

        date_offset = DateOffset()

        # since the targets are in months, this is the value we need.
        first_day_month = start_date.replace(day=1)
        next_date = date_offset.get_offset('1m', first_day_month) - timedelta(days=1)

        target_value = get_target_value(min_date=first_day_month,
                                        max_date=next_date,
                                        target=target,
                                        month_range=True)

        return target_value


class ChartBaseView(ReportBase, ReportUtilsMixin, TemplateView):
    date_field = ReportBuilderDateColumn
    chart_js_table = ChartJSTable

    template_name = 'advanced_report_builder/charts/report.html'

    def __init__(self, *args, **kwargs):
        self.chart_report = None
        self.show_toolbar = False
        self.table = None
        super().__init__(*args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        self.slug = split_slug(kwargs.get('slug'))
        self.enable_edit = kwargs.get('enable_edit')
        self.dashboard_report = kwargs.get('dashboard_report')
        if self.enable_edit or (self.dashboard_report and not self.dashboard_report.top) or not self.dashboard_report:
            self.show_toolbar = True
        return super().dispatch(request, *args, **kwargs)

    def extra_filters(self, query):
        report_query = self.get_report_query(report=self.chart_report)
        if report_query:
            query = self.process_query_filters(query=query,
                                               search_filter_data=report_query.query)
        return query

    def get_date_format(self):
        return '%Y-%m-%d'

    def get_date_field(self, index, fields, base_model, table):

        field_name = self.chart_report.date_field
        if field_name is None:
            return
        report_builder_class = getattr(base_model,
                                       self.chart_report.report_type.report_builder_class_name, None)
        django_field, col_type_override, _, _ = get_field_details(base_model=base_model,
                                                                  field=field_name,
                                                                  report_builder_class=report_builder_class,
                                                                  table=table)

        if col_type_override:
            field_name = col_type_override.field

        date_format = self.get_date_format()

        date_function_kwargs = {'title': field_name,
                                'date_format': date_format}

        annotations_value = self.chart_report.axis_scale

        new_field_name = f'{annotations_value}_{field_name}_{index}'
        function = ANNOTATION_VALUE_FUNCTIONS[annotations_value]
        date_function_kwargs['annotations_value'] = {new_field_name: function(field_name)}
        field_name = new_field_name

        date_function_kwargs.update({'field': field_name,
                                     'column_name': field_name,
                                     'model_path': ''})

        field = self.date_field(**date_function_kwargs)
        fields.append(field)
        return field_name

    @staticmethod
    def _set_multiple_title(database_values, value_prefix, fields, text):
        results = {}
        for field in fields:
            value = database_values[value_prefix + '__' + field]
            results[field] = value
        return text.format(**results)

    def process_query_results(self, base_model, table):
        fields = []
        date_field_name = self.get_date_field(0, fields, base_model=base_model, table=table)
        if date_field_name is not None:
            table.order_by = [date_field_name]
        else:
            table.order_by = ['id']

        if not self.chart_report.fields:
            return fields
        chart_fields = self.chart_report.fields

        report_builder_class = getattr(base_model,
                                       self.chart_report.report_type.report_builder_class_name, None)

        for index, table_field in enumerate(chart_fields, 1):
            field = table_field['field']

            django_field, col_type_override, _, _ = get_field_details(base_model=base_model,
                                                                      field=field,
                                                                      report_builder_class=report_builder_class,
                                                                      table=table)

            if (isinstance(django_field, NUMBER_FIELDS) or
                    (col_type_override is not None and col_type_override.annotations)):
                data_attr = split_attr(table_field)
                if data_attr.get('multiple_columns') == '1':
                    query = self.extra_filters(query=table.model.objects)
                    multiple_column_field = data_attr.get('multiple_column_field')
                    report_builder_class = self._get_report_builder_class(base_model=base_model,
                                                                          field_str=multiple_column_field,
                                                                          report_builder_class=report_builder_class)
                    _fields = report_builder_class.default_multiple_column_fields
                    default_multiple_column_fields = [multiple_column_field + '__' + x for x in _fields]
                    order_by = f'{multiple_column_field}__{report_builder_class.default_multiple_pk}'
                    results = query.order_by(order_by).distinct(multiple_column_field).\
                        values(multiple_column_field, *default_multiple_column_fields)

                    for multiple_index, result in enumerate(results):
                        suffix = self._set_multiple_title(database_values=result,
                                                          value_prefix=multiple_column_field,
                                                          fields=_fields,
                                                          text=report_builder_class.default_multiple_column_text)
                        extra_filter = Q((multiple_column_field, result[multiple_column_field]))

                        self.get_number_field(annotations_type=self.chart_report.axis_value_type,
                                              index=f'{index}_{multiple_index}',
                                              data_attr=data_attr,
                                              table_field=table_field,
                                              fields=fields,
                                              col_type_override=col_type_override,
                                              extra_filter=extra_filter,
                                              title_suffix=suffix,
                                              multiple_index=multiple_index)

                else:
                    self.get_number_field(annotations_type=self.chart_report.axis_value_type,
                                          index=index,
                                          data_attr=data_attr,
                                          table_field=table_field,
                                          fields=fields,
                                          col_type_override=col_type_override)
        return fields

    # noinspection PyUnresolvedReferences
    @staticmethod
    def add_colour_offset(colour, multiple_index):
        if multiple_index > 0:
            offset = 50 * multiple_index
            colour_list = list(int(colour[i:i + 2], 16) for i in (0, 2, 4))
            _, colour_list[0] = divmod(colour_list[0] + offset, 255)
            _, colour_list[1] = divmod(colour_list[1] + offset, 255)
            _, colour_list[2] = divmod(colour_list[2] + offset, 255)
            return "{:02x}{:02x}{:02x}".format(*colour_list)
        else:
            return colour

    def setup_table(self, base_model):
        if hasattr(self.chart_report, 'axis_scale'):
            self.table = self.chart_js_table(model=base_model, axis_scale=self.chart_report.axis_scale)
        else:
            self.table = self.chart_js_table(model=base_model)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        base_model = self.chart_report.get_base_modal()

        self.setup_table(base_model=base_model)
        self.table.extra_filters = self.extra_filters
        fields = self.process_query_results(base_model=base_model, table=self.table)
        self.table.add_columns(*fields)
        context['show_toolbar'] = self.show_toolbar
        context['datatable'] = self.table
        context['title'] = self.get_title()
        return context

    def setup_menu(self):
        if not self.show_toolbar:
            return

        if self.dashboard_report and self.enable_edit:
            report_menu = self.pod_dashboard_edit_menu()
        elif self.dashboard_report and not self.enable_edit:
            report_menu = self.pod_dashboard_view_menu()
        else:
            report_menu = self.pod_report_menu()

        self.add_menu('button_menu', 'button_group').add_items(
            *report_menu,
            *self.queries_menu(),
        )

    def pod_dashboard_edit_menu(self):
        return [MenuItem(f'advanced_report_builder:dashboard_report_modal,pk-{self.dashboard_report.id}',
                         menu_display='Edit',
                         font_awesome='fas fa-pencil-alt', css_classes=['btn-primary'])]

    # noinspection PyMethodMayBeStatic
    def pod_dashboard_view_menu(self):
        return []

    def pod_report_menu(self):
        query_id = self.slug.get(f'query{self.chart_report.id}')
        slug_str = ''
        if query_id:
            slug_str = f'-query_id-{query_id}'
        return self.edit_report_menu(request=self.request, chart_report_id=self.chart_report.id, slug_str=slug_str)

    @staticmethod
    def edit_report_menu(request, chart_report_id, slug_str):
        return []

    # noinspection PyMethodMayBeStatic
    def queries_menu(self):
        return []


class ChartBaseFieldForm(CrispyForm):
    cancel_class = 'btn-secondary modal-cancel'

    def __init__(self, *args, **kwargs):
        self.django_field = None
        self.col_type_override = None
        super().__init__(*args, **kwargs)

    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        if isinstance(self.django_field, NUMBER_FIELDS):
            return StrictButton(button_text, onclick=f'save_modal_{self.form_id}()', css_class=css_class, **kwargs)
        else:
            return super().submit_button(css_class, button_text, **kwargs)

    def cancel_button(self, css_class=cancel_class, **kwargs):
        commands = [{'function': 'save_query_builder_id_query_data'},
                    {'function': 'close'}]
        return self.button('Cancel', commands, css_class, **kwargs)

    def get_report_type_details(self):
        data = json.loads(base64.b64decode(self.slug['data']))

        report_type = get_object_or_404(ReportType, pk=self.slug['report_type_id'])
        base_model = report_type.content_type.model_class()
        report_builder_class = getattr(base_model,
                                       report_type.report_builder_class_name, None)
        self.django_field, self.col_type_override, _, _ = get_field_details(base_model=base_model,
                                                                            field=data['field'],
                                                                            report_builder_class=report_builder_class)

        return report_type, base_model

    def _get_query_builder_foreign_key_fields(self, base_model, report_builder_fields, fields,
                                              prefix='', title_prefix='', previous_base_model=None):
        for include_field, include in report_builder_fields.includes.items():
            app_label, model, report_builder_fields_str = include['model'].split('.')
            new_model = apps.get_model(app_label, model)
            if new_model != previous_base_model:
                new_report_builder_fields = getattr(new_model, report_builder_fields_str, None)
                fields.append((prefix + include_field, title_prefix + include['title']))
                self._get_query_builder_foreign_key_fields(base_model=new_model,
                                                           report_builder_fields=new_report_builder_fields,
                                                           fields=fields,
                                                           prefix=f"{prefix}{include_field}__",
                                                           title_prefix=f"{title_prefix}{include['title']} --> ",
                                                           previous_base_model=base_model)



from django.db import models
from django_datatables.columns import ManyToManyColumn

from advanced_report_builder.columns import FilterForeignKeyColumn
from advanced_report_builder.globals import DATE_FIELDS
from advanced_report_builder.variable_date import VariableDate


class FieldTypes:
    FIELD_TYPE_STRING = 1
    FIELD_TYPE_NUMBER = 2
    FIELD_TYPE_DATE = 3
    FIELD_TYPE_BOOLEAN = 4
    FIELD_TYPE_MULTIPLE_CHOICE = 5
    FIELD_TYPE_FOREIGN_KEY = 6
    FIELD_TYPE_ABSTRACT_USER = 7

    def get_operator(self, field_type):
        operators = {self.FIELD_TYPE_STRING: ['equal',
                                              'not_equal',
                                              'contains',
                                              'not_contains',
                                              'begins_with',
                                              'not_begins_with',
                                              'ends_with',
                                              'not_ends_with'
                                              ],
                     self.FIELD_TYPE_NUMBER: ['equal',
                                              'not_equal',
                                              'less',
                                              'less_or_equal',
                                              'greater',
                                              'greater_or_equal'
                                              ],
                     self.FIELD_TYPE_DATE: ['equal',
                                            'not_equal',
                                            'less',
                                            'less_or_equal',
                                            'greater',
                                            'greater_or_equal',
                                            'is_null',
                                            'is_not_null'
                                            ],
                     self.FIELD_TYPE_BOOLEAN: ['equal',
                                               'not_equal'
                                               ],
                     self.FIELD_TYPE_MULTIPLE_CHOICE: ['in',
                                                       'not_in'],
                     self.FIELD_TYPE_FOREIGN_KEY: ['is_null',
                                                   'is_not_null'],
                     self.FIELD_TYPE_ABSTRACT_USER: ['equal',
                                                     'not_equal'],
                     }
        return operators.get(field_type)

    def get_filter(self, query_builder_filters, django_field, field, title, column, prefix):
        if isinstance(column.field, list):
            full_field_name = prefix + column.field[0]
        elif column.field is not None:
            full_field_name = prefix + column.field
        else:
            full_field_name = prefix + field
        column_id = prefix + field
        if django_field is not None:
            if isinstance(column, FilterForeignKeyColumn):
                query_builder_filters.append({"id": column_id,
                                              "label": title,
                                              "field": full_field_name,
                                              "type": "string",
                                              'input': 'select',
                                              'multiple': True,
                                              'values': column.get_query_options(),
                                              "operators": self.get_operator(
                                                  self.FIELD_TYPE_MULTIPLE_CHOICE),
                                              })
            elif isinstance(django_field, (models.CharField, models.TextField, models.EmailField)):
                query_builder_filters.append({"id": column_id,
                                              "label": title,
                                              "field": full_field_name,
                                              "type": "string",
                                              "operators": self.get_operator(self.FIELD_TYPE_STRING),
                                              "validation": {"allow_empty_value": True
                                                             }
                                              })
            elif isinstance(django_field, (models.IntegerField,
                                           models.PositiveSmallIntegerField,
                                           models.PositiveIntegerField)):
                if django_field.choices is None:
                    query_builder_filter = {"id": column_id,
                                            "label": title,
                                            "field": full_field_name,
                                            "type": "integer",
                                            "operators": self.get_operator(self.FIELD_TYPE_NUMBER),
                                            }
                else:
                    query_builder_filter = {"id": column_id,
                                            "label": title,
                                            "field": full_field_name,
                                            "type": "integer",
                                            'input': 'select',
                                            'multiple': True,
                                            'values': dict(django_field.choices),
                                            "operators": self.get_operator(self.FIELD_TYPE_MULTIPLE_CHOICE),
                                            }
                query_builder_filters.append(query_builder_filter)
            elif isinstance(django_field, models.BooleanField):
                query_builder_filters.append({"id": column_id,
                                              "label": title,
                                              "field": full_field_name,
                                              "input": "select",
                                              "operators": self.get_operator(self.FIELD_TYPE_BOOLEAN),
                                              "values": {"0": "False", "1": "True"}})
            elif isinstance(django_field, DATE_FIELDS):
                self.get_date_field(column_id=column_id,
                                    query_builder_filters=query_builder_filters,
                                    field=full_field_name,
                                    title=title)
        else:
            if isinstance(column, ManyToManyColumn):
                choices = dict(column.options['lookup'])
                query_builder_filter = {"id": column_id,
                                        "label": title,
                                        "field": prefix + column.field_id,
                                        "type": "integer",
                                        'input': 'select',
                                        'multiple': True,
                                        'values': choices,
                                        "operators": self.get_operator(self.FIELD_TYPE_MULTIPLE_CHOICE)}
                query_builder_filters.append(query_builder_filter)

    def get_foreign_key_null_field(self, query_builder_filters, field, title):
        query_builder_filters.append({"id": field,
                                      "label": title,
                                      "field": field,
                                      "type": "integer",
                                      "operators": self.get_operator(self.FIELD_TYPE_FOREIGN_KEY),
                                      "validation": {"allow_empty_value": True
                                                     }
                                      })

    def get_date_field(self, column_id, query_builder_filters, field, title):
        variable_date = VariableDate()
        query_builder_filter = {"id": f'{column_id}__variable_date',
                                "label": f'{title} (Variable)',
                                "field": field,
                                "operators": self.get_operator(self.FIELD_TYPE_DATE),
                                "input": "select",
                                "values": variable_date.get_variable_date_filter_values()
                                }
        query_builder_filters.append(query_builder_filter)

    def get_abstract_user_field(self, query_builder_filters, field, title):
        query_builder_filters.append({"id": f'{field}__logged_in_user',
                                      "label": f'{title} (Logged in user)',
                                      "field": field,
                                      "input": "select",
                                      "operators": self.get_operator(self.FIELD_TYPE_ABSTRACT_USER),
                                      "values": {"0": "False", "1": "True"}
                                      })



# version = pip_version('django-advanced-report-builder')
from ajax_helpers.html_include import SourceBase


class Dot(SourceBase):
    static_path = '/advanced_report_builder/dot/'
    js_filename = 'doT.js'


class JQueryExtendext(SourceBase):
    static_path = '/advanced_report_builder/jquery_extendext/'
    js_filename = 'jQuery.extendext.js'


class QueryBuilder(SourceBase):
    static_path = '/advanced_report_builder/query_builder/'
    css_filename = 'query-builder.default.css'
    js_filename = 'query-builder.min.js'


class DashboardInclude(SourceBase):
    static_path = '/advanced_report_builder/dashboard/'
    js_filename = 'dashboard.js'


class ChartJS(SourceBase):
    static_path = '/advanced_report_builder/chart-js/'
    js_filename = ['chart.min.js',
                   'chartjs-adapter-moment.min.js',  # this does require moment however it's already included
                   'chartjs-plugin-datalabels.min.js']


class D3(SourceBase):
    static_path = '/advanced_report_builder/d3/'
    js_filename = ['d3.min.js']


class D3Funnel(SourceBase):
    static_path = '/advanced_report_builder/d3-funnel/'
    js_filename = ['d3-funnel.min.js']


packages = {
    'query_builder': [JQueryExtendext, Dot, QueryBuilder],
    'dashboard': [DashboardInclude],
    'funnel': [D3, D3Funnel],
}



import operator
from functools import reduce

from django.apps import apps
from django.db.models import Q
from django.shortcuts import get_object_or_404

from advanced_report_builder.models import ReportQuery
from advanced_report_builder.variable_date import VariableDate


class FilterQueryMixin:

    def __init__(self, *args, **kwargs):
        self.report = None
        self.dashboard_report = None
        self.enable_edit = None
        self.slug = None
        self.base_model = None
        super().__init__(*args, **kwargs)

    def process_query_filters(self, query, search_filter_data):
        result = self.process_filters(search_filter_data=search_filter_data)
        if result:
            return query.filter(result)
        return query

    def process_filters(self, search_filter_data, extra_filter=None):
        if not search_filter_data:
            return []

        query_list = self._process_group(query_data=search_filter_data)
        if extra_filter:
            query_list.append(extra_filter)

        reduce_by = self._format_group_conditions(search_filter_data['condition'])

        if query_list:
            return reduce(reduce_by, query_list)
        return []

    @staticmethod
    def _get_operator(display_operator):
        """
        Converts the display operators into operators for the Q object.
        :param display_operator:
        :return:
        """

        operators_dict = {
            'equal': '',
            'not_equal': '',  # Not equals is now done by inverting the individual Q object. (for dates).
            'less': '__lt',
            'less_or_equal': '__lte',
            'greater': '__gt',
            'greater_or_equal': '__gte',
            'contains': '__icontains',
            'not_contains': '__icontains',
            'begins_with': '__istartswith',
            'not_begins_with': '__istartswith',
            'ends_with': '__iendswith',
            'not_ends_with': '__iendswith',
            'in': '__in',
            'not_in': '__in',
            'is_null': '__isnull',
            'is_not_null': '__isnull',
        }

        query_operator = operators_dict.get(display_operator)
        return query_operator

    @staticmethod
    def _format_group_conditions(display_condition):
        """
        Converts the display condition from the javascript into a condition for the Q object.
        :param display_condition:
        :return:
        """
        if display_condition == 'AND':
            reduce_by = operator.and_
        else:
            reduce_by = operator.or_

        return reduce_by

    def _process_group(self, query_data):
        query_list = []

        for rule in query_data['rules']:
            if condition := rule.get('condition'):
                reduce_by = self._format_group_conditions(condition)
                sub_query_list = self._process_group(query_data=rule)
                if sub_query_list:
                    query_list.append(reduce(reduce_by, sub_query_list))
                continue

            field = rule['field']
            _id = rule['id']
            display_operator = rule['operator']
            query_string = field + self._get_operator(display_operator)
            data_type = rule['type']
            value = rule.get('value', '')
            if display_operator == 'is_null':
                value = True
            elif display_operator == 'is_not_null':
                value = False

            if data_type == "string" and _id.endswith('__variable_date'):
                self.get_variable_date(value=value,
                                       query_list=query_list,
                                       display_operator=display_operator,
                                       field=field,
                                       query_string=query_string)
            elif data_type == "string" and _id.endswith('__logged_in_user'):
                self.get_logged_in_user(value=value,
                                        query_list=query_list,
                                        display_operator=display_operator,
                                        field=field,
                                        query_string=query_string)
            else:
                # 'Normal' Query.
                if display_operator in ["not_equal", "not_in", "not_contains", 'not_begins_with', "not_ends_with"]:
                    query_list.append(~Q((query_string, value)))
                else:
                    query_list.append(Q((query_string, value)))

        return query_list

    @staticmethod
    def get_variable_date(value, query_list, display_operator, field, query_string):
        if display_operator in ['is_null', 'is_not_null']:
            query_list.append(Q((query_string, value)))
        else:
            _, range_type = value.split(":")
            variable_date = VariableDate()
            value = variable_date.get_variable_dates(range_type=int(range_type))
            if display_operator in ['less', 'greater_or_equal']:
                query_list.append(Q((query_string, value[0])))
            elif display_operator in ['greater', 'less_or_equal']:
                query_list.append(Q((query_string, value[1])))
            elif display_operator in ["not_equal", "not_in"]:
                query_list.append(~((Q((field + "__gte", value[0]))) & (Q((field + "__lte", value[1])))))
            else:
                query_list.append(((Q((field + "__gte", value[0]))) & (Q((field + "__lte", value[1])))))

    def get_logged_in_user(self, value, query_list, display_operator, field, query_string):

        # noinspection PyUnresolvedReferences
        current_user = self.request.user
        value = value == '1'
        if (display_operator == 'equal' and value) or (display_operator == 'not_equal' and not value):
            query_list.append(Q((query_string, current_user)))
        else:
            query_list.append(~Q((query_string, current_user)))

    def get_report_query(self, report):

        slug_key = self.slug.get(f'query{report.pk}')
        if slug_key:
            report_query = get_object_or_404(ReportQuery, id=slug_key)
            if report_query.report_id != report.pk:
                return None
        else:
            report_query = report.reportquery_set.first()
        return report_query

    def get_title(self):
        if self.dashboard_report and self.dashboard_report.name_override:
            return self.dashboard_report.name_override
        else:
            return self.report.name

    def _get_report_builder_class(self, base_model, field_str, report_builder_class, previous_base_model=None):
        if '__' in field_str:
            field_parts = field_str.split('__')
            include_str = field_parts[0]
            new_field_str = '__'.join(field_parts[1:])

            include = report_builder_class.includes.get(include_str)

            if include is not None:
                app_label, model, report_builder_fields_str = include['model'].split('.')
                new_model = apps.get_model(app_label, model)
                new_report_builder_fields = getattr(new_model, report_builder_fields_str, None)

                if new_model != previous_base_model:
                    result = self._get_report_builder_class(base_model=new_model,
                                                            field_str=new_field_str,
                                                            report_builder_class=new_report_builder_fields,
                                                            previous_base_model=base_model)
                    if result:
                        return result
        else:
            include = report_builder_class.includes.get(field_str)
            if include is not None:
                app_label, model, report_builder_fields_str = include['model'].split('.')
                new_model = apps.get_model(app_label, model)
                new_report_builder_fields = getattr(new_model, report_builder_fields_str, None)
                return new_report_builder_fields
        return None

    def _get_pivot_details(self, base_model, pivot_str, report_builder_class, previous_base_model=None, include_str=''):

        if pivot_str in report_builder_class.pivot_fields:
            pivot_data = report_builder_class.pivot_fields[pivot_str]
            if include_str == '':
                full_field_id = pivot_data['field']
            else:
                full_field_id = '__'.join((include_str, pivot_data['field']))

            return {'id': full_field_id,
                    'details': pivot_data}

        if '__' in pivot_str:
            pivot_parts = pivot_str.split('__')
            include_str = pivot_parts[0]
            new_pivot_str = '__'.join(pivot_parts[1:])

            include = report_builder_class.includes.get(include_str)
            if include is not None:
                app_label, model, report_builder_fields_str = include['model'].split('.')
                new_model = apps.get_model(app_label, model)
                new_report_builder_fields = getattr(new_model, report_builder_fields_str, None)

                if new_model != previous_base_model:
                    result = self._get_pivot_details(base_model=new_model,
                                                     pivot_str=new_pivot_str,
                                                     report_builder_class=new_report_builder_fields,
                                                     previous_base_model=base_model,
                                                     include_str=include_str)
                    if result:
                        return result
        return None



from django.forms.widgets import Input


class DataMergeWidget(Input):
    template_name = 'advanced_report_builder/data_merge/data_merge.html'
    input_type = 'textarea'
    crispy_kwargs = {'label_class': 'col-3 col-form-label-sm', 'field_class': 'col-12 input-group-sm'}

    def __init__(self, data_merge_data=None):
        super().__init__()
        self.data_merge_data = data_merge_data

    def get_context(self, name, value, attrs=None):
        context = super().get_context(name, value, attrs)
        if context['widget']['value'] is None:
            context['widget']['value'] = ''
        if self.data_merge_data is None:
            context['data_merge'] = None
        else:
            context['data_merge'] = self.data_merge_data.build_menus()

        return context



from django.contrib.humanize.templatetags.humanize import intcomma
from django.db.models import Count
from django_datatables.columns import ColumnBase, CurrencyPenceColumn, CurrencyColumn, NoHeadingColumn, ColumnLink
from django_datatables.helpers import get_url, DUMMY_ID


class ReportBuilderDateColumn(ColumnBase):

    def __init__(self, *,  date_format=None, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        if date_format:
            self.date_format = date_format
        else:
            self.date_format = '%d/%m/%Y'

    def row_result(self, data, _page_data):
        try:
            date = data[self.field].strftime(self.date_format)
            return date
        except AttributeError:
            return ""


class ReportBuilderNumberColumn(ColumnBase):

    def __init__(self, *,  decimal_places=0, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        self.decimal_places = f'{{:.{decimal_places}f}}'

    def row_result(self, data, _page_data):
        number = data.get(self.field)
        if number is None:
            return ''
        else:
            number = self.decimal_places.format(data[self.field])
        if '.' in number:
            return number.rstrip('0').rstrip('.')
        else:
            return number


class ReportBuilderCurrencyPenceColumn(CurrencyPenceColumn):

    def row_result(self, data, _page_data):
        try:
            return intcomma('{:.2f}'.format(data[self.field] / 100.0))
        except (KeyError, TypeError):
            return '0.00'


class ReportBuilderCurrencyColumn(CurrencyColumn):

    def row_result(self, data, _page_data):
        try:
            return intcomma('{:.2f}'.format(data[self.field]))
        except (KeyError, TypeError):
            return '0.00'


class ArrowColumn(NoHeadingColumn):
    def __init__(self, **kwargs):
        kwargs['render'] = [{'html': '<i class="fa fa-angle-right fa-2x"></i>', 'function': 'Html'}]
        kwargs['width'] = '10px'
        super().__init__(**kwargs)


class ColourColumn(ColumnBase):
    def row_result(self, data, _page_data):
        colour = data.get(self.field)
        if colour is not None:
            return '#' + colour
        return ''


class FilterForeignKeyColumn(ColumnBase):
    def get_query_options(self):
        values = self.model.objects.distinct(self.field).order_by(self.field).values_list(self.field, flat=True)
        return {v: v for v in values if v}


class ReportBuilderColumnLink(ColumnLink):

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url_name):
        if not self.table or self.table.view.kwargs.get('enable_links'):
            self._url = get_url(url_name)
        else:
            self._url = f'#?{DUMMY_ID}'


class RecordCountColumn(ColumnBase):
    def __init__(self, field=None, **kwargs):
        if 'annotations' not in kwargs:
            kwargs['annotations'] = {'record_count': Count(1)}
        if 'column_name' not in kwargs:
            kwargs['column_name'] = 'record_count'
        super().__init__(field, **kwargs)
